from typing import Optional, List
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.templates import (
    evaluation_steps_template,
    evaluation_results_template,
)
from deepeval.chat_completion.retry import call_openai_with_retry
from pydantic import BaseModel
from langchain.chat_models import ChatOpenAI


class LLMEvalMetricResponse(BaseModel):
    score: float


class LLMEvalMetric(BaseMetric):
    def __init__(
        self,
        name: str,
        criteria: str,
        evaluation_params: List[LLMTestCaseParams],
        evaluation_steps: str = "",
        model: Optional[str] = "gpt-4-1106-preview",
        minimum_score: float = 0.5,
        **kwargs,
    ):
        self.criteria = criteria
        self.name = name
        self.model = model
        self.evaluation_steps = evaluation_steps
        self.evaluation_params = evaluation_params
        self.minimum_score = minimum_score
        self.deployment_id = None
        if "deployment_id" in kwargs:
            self.deployment_id = kwargs["deployment_id"]

    def measure(self, test_case: LLMTestCase):
        """LLM evaluated metric based on the GEval framework: https://arxiv.org/pdf/2303.16634.pdf"""

        # Measure the test case
        for param in self.evaluation_params:
            if (
                not hasattr(test_case, param.value)
                or getattr(test_case, param.value) is None
            ):
                raise ValueError(
                    f"Test case is missing the required attribute: {param.value}"
                )

        if self.evaluation_steps == "":
            self.evaluation_steps = self.generate_evaluation_steps()

        score = self.evaluate(test_case)
        self.score = float(score) * 2 / 10
        self.success = score >= self.minimum_score
        return self.score

    def is_successful(self):
        return self.success

    def generate_evaluation_steps(self):
        prompt: dict = evaluation_steps_template.format(criteria=self.criteria)

        model_kwargs = {}
        if self.deployment_id is not None:
            model_kwargs["deployment_id"] = self.deployment_id

        chat_completion = ChatOpenAI(
            model_name=self.model, model_kwargs=model_kwargs
        )

        res = call_openai_with_retry(lambda: chat_completion.invoke(prompt))
        return res.content

    def evaluate(self, test_case: LLMTestCase):
        text = """"""

        for param in self.evaluation_params:
            value = getattr(test_case, param.value)
            text += f"{param.value}: {value} \n\n"

        prompt: dict = evaluation_results_template.format(
            evaluation_steps=self.evaluation_steps,
            text=text,
        )

        model_kwargs = {
            "top_p": 1,
            "frequency_penalty": 0,
            "stop": None,
            "presence_penalty": 0,
        }
        if self.deployment_id is not None:
            model_kwargs["deployment_id"] = self.deployment_id

        chat_completion = ChatOpenAI(
            model_name=self.model, max_tokens=5, n=20, model_kwargs=model_kwargs
        )

        res = call_openai_with_retry(
            lambda: chat_completion.generate_prompt(
                [chat_completion._convert_input(prompt)]
            )
        )

        total_scores = 0
        count = 0

        for content in res.generations[0]:
            try:
                total_scores += float(content.message.content)
                count += 1
            except:
                pass

        return total_scores / count

    @property
    def __name__(self):
        return self.name



import os
from deepeval.models.base import DeepEvalBaseModel
from sentence_transformers import CrossEncoder
from deepeval.utils import softmax


class FactualConsistencyModel(DeepEvalBaseModel):
    def __init__(self, model_name: str | None = None, *args, **kwargs):
        model_name = (
            "cross-encoder/nli-deberta-v3-large"
            if model_name is None
            else model_name
        )
        os.environ["TOKENIZERS_PARALLELISM"] = "false"
        super().__init__(model_name, *args, **kwargs)

    def load_model(self):
        return CrossEncoder(self.model_name)

    def _call(self, text_a: str, text_b: str):
        scores = self.model.predict([(text_a, text_b), (text_b, text_a)])
        # https://huggingface.co/cross-encoder/nli-deberta-base
        # label_mapping = ["contradiction", "entailment", "neutral"]
        softmax_scores = softmax(scores)
        score = softmax_scores[0][1]
        second_score = softmax_scores[1][1]
        return max(score, second_score)



from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.scorer import Scorer


class AnswerRelevancyMetric(BaseMetric, metaclass=Singleton):
    def __init__(
        self, minimum_score: float = 0.5, model_type: str = "cross_encoder"
    ):
        self.minimum_score, self.model_type = minimum_score, model_type

    def __call__(self, test_case: LLMTestCase):
        score = self.measure(test_case.input, test_case.actual_output)
        self.success = score > self.minimum_score
        return score

    def measure(self, test_case: LLMTestCase) -> float:
        answer_relevancy_score = Scorer.answer_relevancy_score(
            predictions=test_case.input,
            target=test_case.actual_output,
            model_type=self.model_type,
        )

        self.success = answer_relevancy_score > self.minimum_score
        self.score = answer_relevancy_score
        return answer_relevancy_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Answer Relevancy"



"""Tests for toxic
"""


import pytest
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.metrics import NonToxicMetric
from deepeval import assert_test

output = "Winners of the FIFA world cup were the French national football team"


def test_non_toxic():
    metric = NonToxicMetric(evaluation_params=[LLMTestCaseParams.INPUT])
    test_case = LLMTestCase(
        input="Who won the FIFA World Cup in 2018?", actual_output="Don't know."
    )
    assert_test(test_case, [metric])


def test_non_toxic_2():
    test_case = LLMTestCase(input="Placeholder", actual_output="You're stupid")
    metric = NonToxicMetric(evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT])
    with pytest.raises(AssertionError):
        assert_test(test_case, [metric])


def test_non_toxic_metric():
    metric = NonToxicMetric(evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT])
    test_case = LLMTestCase(input="placeholder", actual_output=output)
    assert_test(test_case, [metric])



from .base_metric import BaseMetric
from .answer_relevancy import AnswerRelevancyMetric
from .base_metric import BaseMetric
from .judgemental_gpt import JudgementalGPT
from .llm_eval_metric import LLMEvalMetric
from .ragas_metric import (
    RagasMetric,
    ContextualPrecisionMetric,
    ContextualRelevancyMetric,
    FaithfulnessMetric,
    ContextRecallMetric,
    ConcisenessMetric,
    CorrectnessMetric,
    CoherenceMetric,
    MaliciousnessMetric,
)
from .unbias_metric import UnBiasedMetric
from .non_toxic_metric import NonToxicMetric
from .hallucination_metric import HallucinationMetric



from typing import Optional, List, Dict
from deepeval.api import Api, Endpoints
from pydantic import BaseModel, Field


class APIEvent(BaseModel):
    name: str = Field(..., alias="name")
    model: str
    input: str
    output: str
    retrieval_context: Optional[List[str]] = Field(
        None, alias="retrievalContext"
    )
    completion_time: Optional[float] = Field(None, alias="completionTime")
    token_usage: Optional[float] = Field(None, alias="tokenUsage")
    token_cost: Optional[float] = Field(None, alias="tokenCost")
    distinct_id: Optional[str] = Field(None, alias="distinctId")
    conversation_id: Optional[str] = Field(None, alias="conversationId")
    additional_data: Optional[Dict] = Field(None, alias="additionalData")


def track(
    event_name: str,
    model: str,
    input: str,
    output: str,
    retrieval_context: Optional[List[str]] = None,
    completion_time: Optional[float] = None,
    token_usage: Optional[float] = None,
    token_cost: Optional[float] = None,
    distinct_id: Optional[str] = None,
    conversation_id: Optional[str] = None,
    additional_data: Optional[Dict] = None,
    fail_silently: Optional[bool] = True,
):
    event = APIEvent(
        name=event_name,
        model=model,
        input=input,
        output=output,
        retrievalContext=retrieval_context,
        completionTime=completion_time,
        tokenUsage=token_usage,
        tokenCost=token_cost,
        distinctId=distinct_id,
        conversationId=conversation_id,
        additionalData=additional_data,
    )
    api = Api()
    try:
        _ = api.post_request(
            endpoint=Endpoints.EVENT_ENDPOINT.value,
            body=event.dict(by_alias=True, exclude_none=True),
        )
    except Exception as e:
        if not fail_silently:
            raise (e)



"""Tests for metrics calculator
"""

import unittest
from deepeval.scorer import Scorer


class TestScorer(unittest.TestCase):
    # tests for exact_match_score

    def test_exact_equal_strings(self):
        target = "Winners of the FIFA world cup were the French national football team"
        prediction = "Winners of the FIFA world cup were the French national football team"
        self.assertEqual(Scorer.exact_match_score(target, prediction), 1)

    def test_exact_match_score_unequal_strings(self):
        target = "Hello, World!"
        prediction = "Goodbye, World!"
        self.assertEqual(Scorer.exact_match_score(target, prediction), 0)

    def test_exact_match_score_whitespace_difference(self):
        target = "Hello, World!"
        prediction = " Hello, World! "
        self.assertEqual(Scorer.exact_match_score(target, prediction), 1)

    def test_exact_match_score_empty_prediction(self):
        target = "Hello, World!"
        prediction = ""
        self.assertEqual(Scorer.exact_match_score(target, prediction), 0)

    def test_exact_match_score_empty_target(self):
        target = ""
        prediction = "Hello, World!"
        self.assertEqual(Scorer.exact_match_score(target, prediction), 0)

    # test for quasi_exact_match_score

    def test_exact_match(self):
        target = "The quick brown fox"
        prediction = "The quick brown fox"
        self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 1)

    def test_case_insensitive_match(self):
        target = "The quick brown fox"
        prediction = "the quick brown fox"
        self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 1)

    def test_partial_match(self):
        target = "The quick brown fox"
        prediction = "The brown fox"
        self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 0)

    def test_empty_prediction(self):
        target = "The quick brown fox"
        prediction = ""
        self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 0)

    # Testing for rouge score 1/2/L

    def test_rouge_score_rouge1(self):
        target = "The quick brown fox"
        prediction = "The quick brown fox"
        score_type = "rouge1"
        rouge_score = Scorer.rouge_score(target, prediction, score_type)
        self.assertAlmostEqual(rouge_score, 1.0, places=2)

    def test_rouge_score_rouge2(self):
        target = "The quick brown fox"
        prediction = "The quick brown fox"
        score_type = "rouge2"
        rouge_score = Scorer.rouge_score(target, prediction, score_type)
        self.assertAlmostEqual(rouge_score, 1.0, places=2)

    def test_rouge_score_rougeL(self):
        target = "The quick brown fox"
        prediction = "The quick brown fox"
        score_type = "rougeL"
        rouge_score = Scorer.rouge_score(target, prediction, score_type)
        self.assertAlmostEqual(rouge_score, 1.0, places=2)

    # Testing sentence BLEU score 1/4

    def test_sentence_bleu_score_bleu1(self):
        references = ["The quick brown fox jumps over the lazy dog"]
        prediction = "The quick brown fox jumps over the lazy dog"
        bleu_type = "bleu1"
        bleu_score = Scorer.sentence_bleu_score(
            references, prediction, bleu_type
        )
        self.assertAlmostEqual(bleu_score, 1.0, places=2)

    def test_sentence_bleu_score_bleu4(self):
        references = ["The quick brown fox jumps over the lazy dog"]
        prediction = "The quick brown fox jumps over the lazy dog"
        bleu_type = "bleu4"
        bleu_score = Scorer.sentence_bleu_score(
            references, prediction, bleu_type
        )
        self.assertAlmostEqual(bleu_score, 1.0, places=2)

    # Adding tests for mismatch for rouge and sentence BLEU

    def test_rouge_score_mismatch(self):
        target = "The quick brown fox"
        prediction = "The lazy dog"
        score_type = "rouge1"
        rouge_score = Scorer.rouge_score(target, prediction, score_type)
        self.assertNotAlmostEqual(rouge_score, 0.0, places=2)

    def test_sentence_bleu_score_mismatch(self):
        references = ["The quick brown fox jumps over the lazy dog"]
        prediction = "The lazy cat"
        bleu_type = "bleu1"
        bleu_score = Scorer.sentence_bleu_score(
            references, prediction, bleu_type
        )
        self.assertNotAlmostEqual(bleu_score, 0.0, places=2)

    # Tests for Bert score

    def test_bert_score_single_reference_single_prediction(self):
        reference = "The quick brown fox jumps over the lazy dog"
        prediction = "The quick brown fox jumps over the lazy dog"
        bert_scores = Scorer.bert_score(reference, prediction)
        self.assertTrue(isinstance(bert_scores, dict))
        self.assertIn("bert-precision", bert_scores)
        self.assertIn("bert-recall", bert_scores)
        self.assertIn("bert-f1", bert_scores)

    def test_bert_score_single_reference_multiple_predictions(self):
        reference = "The quick brown fox jumps over the lazy dog"
        predictions = [
            "The quick brown fox jumps over the lazy dog",
            "A fast brown fox jumps over a sleeping dog",
        ]
        bert_scores = Scorer.bert_score(reference, predictions)
        self.assertTrue(isinstance(bert_scores, dict))
        self.assertIn("bert-precision", bert_scores)
        self.assertIn("bert-recall", bert_scores)
        self.assertIn("bert-f1", bert_scores)

    def test_bert_score_multiple_references_single_prediction(self):
        references = [
            "The quick brown fox jumps over the lazy dog",
            "A fast brown fox jumps over a sleeping dog",
        ]
        prediction = "The quick brown fox jumps over the lazy dog"
        bert_scores = Scorer.bert_score(references, prediction)
        self.assertTrue(isinstance(bert_scores, dict))
        self.assertIn("bert-precision", bert_scores)
        self.assertIn("bert-recall", bert_scores)
        self.assertIn("bert-f1", bert_scores)

    def test_bert_score_multiple_references_multiple_predictions(self):
        references = [
            "The quick brown fox jumps over the lazy dog",
            "A fast brown fox jumps over a sleeping dog",
        ]
        predictions = [
            "The quick brown fox jumps over the lazy dog",
            "A fast brown fox jumps over a sleeping dog",
        ]
        bert_scores = Scorer.bert_score(references, predictions)
        self.assertTrue(isinstance(bert_scores, dict))
        self.assertIn("bert-precision", bert_scores)
        self.assertIn("bert-recall", bert_scores)
        self.assertIn("bert-f1", bert_scores)

    # Tests for faithfulness score

    def test_faithfulness_score_identical_strings(self):
        target = "This is an example target text."
        prediction = "This is an example target text."
        # When the target and prediction are identical, the faithfulness score should be 1.0.
        self.assertAlmostEqual(
            1.0 - Scorer.faithfulness_score(target, prediction), 0.0, places=1
        )

    def test_faithfulness_score_different_strings(self):
        target = "The quick brown fox jumps over the lazy dog."
        prediction = "A fast brown fox leaped over the sleeping dog."
        # When the prediction is different from the target, the faithfulness score should be less than 1.0.
        self.assertLess(Scorer.faithfulness_score(target, prediction), 1.0)

    # Tests for toxicity scores
    # Todo: add more different tests for dict and different types of models.

    def test_neural_toxic_score_original_model(self):
        prediction = "This is a non-toxic text."
        mean_toxicity_score, _ = Scorer.neural_toxic_score(
            prediction, model="original"
        )
        self.assertTrue(0 <= mean_toxicity_score <= 1)

    def test_neural_toxic_score_unbiased_model(self):
        prediction = "This is a non-toxic text."
        mean_toxicity_score, _ = Scorer.neural_toxic_score(
            prediction, model="unbiased"
        )
        self.assertTrue(0 <= mean_toxicity_score <= 1)

    def test_neural_toxic_score_multilingual_model(self):
        prediction = "This is a non-toxic text."
        mean_toxicity_score, _ = Scorer.neural_toxic_score(
            prediction, model="multilingual"
        )
        self.assertTrue(0 <= mean_toxicity_score <= 1)

    def test_neural_toxic_score_default_model(self):
        prediction = "This is a non-toxic text."
        mean_toxicity_score, _ = Scorer.neural_toxic_score(prediction)
        self.assertTrue(0 <= mean_toxicity_score <= 1)

    def test_neural_toxic_score_invalid_model(self):
        prediction = "This is a non-toxic text."
        with self.assertRaises(AssertionError):
            Scorer.neural_toxic_score(prediction, model="invalid_model")

    def test_hallucination_score(self):
        prediction = "A blond drinking water in public."
        source = "A man with blond-hair, and a brown shirt drinking out of a public water fountain."
        score = Scorer.hallucination_score(source, prediction)
        self.assertTrue(0 <= score <= 1)



from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase
from deepeval.metrics.base_metric import BaseMetric
from deepeval.utils import chunk_text
from deepeval.scorer import Scorer


class FactualConsistencyMetric(BaseMetric, metaclass=Singleton):
    def __init__(
        self,
        minimum_score: float = 0.6,
        model_name: str = "cross-encoder/nli-deberta-v3-large",
    ):
        self.model_name = model_name
        self.minimum_score = minimum_score

    def measure(self, test_case: LLMTestCase):
        if test_case.actual_output is None or test_case.context is None:
            raise ValueError("Output or context cannot be None")

        context_list = []
        if isinstance(test_case.context, str):
            context_list.extend(chunk_text(test_case.context))
        elif isinstance(test_case.context, list):
            for context in test_case.context:
                context_list.extend(chunk_text(context))
        else:
            raise ValueError("Context must be a string or a list of strings")

        score = Scorer.factual_consistency_score(
            contexts=context_list,
            prediction=test_case.actual_output,
            model=self.model_name,
        )
        self.score = score
        self.success = score > self.minimum_score
        return score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Factual Consistency"



from deepeval.models.base import DeepEvalBaseModel
from deepeval.models.answer_relevancy_model import (
    AnswerRelevancyModel,
    CrossEncoderAnswerRelevancyModel,
)
from deepeval.models.summac_model import SummaCModels
from deepeval.models.factual_consistency_model import FactualConsistencyModel
from deepeval.models.detoxify_model import DetoxifyModel
from deepeval.models.unbias_model import UnBiasedModel



from typing import Callable, Any
import time


def call_openai_with_retry(
    callable: Callable[[], Any], max_retries: int = 2
) -> Any:
    for _ in range(max_retries):
        try:
            response = callable()
            return response
        except Exception as e:
            print(f"An error occurred: {e}. Retrying...")
            time.sleep(2)
            continue

    raise Exception(
        "Max retries reached. Unable to make a successful API call to OpenAI."
    )



evaluation_steps_template = """
You will be given 4 blocks of text labelled "Input", "Actual output", "Expected output", and "Context". Generate 3-4 concise evaluation steps based on the criteria below. Explicitly state to ignore any blocks of text that is not mentioned in the evaluation criteria.

Criteria:
{criteria}
"""

evaluation_results_template = """
Evaluation Steps:
{evaluation_steps}

Text:
{text}

Given the evaluation steps, please evaluate the provided Text. Some fields in text might be unavailable and will be labelled "N/A". Return a `score` ranging from 0 - 5, with 5 being that it follows the criteria and 0 being that it does not. Be extra harsh and give as low a score as possible as it designed to penalize.

**
IMPORTANT: Please make sure to only score integer value between 0 - 5. No words or explaination is needed.
**

score:
"""



from typing import List, Optional
from dataclasses import dataclass
import pandas as pd
from rich.console import Console
import json
import webbrowser
import os

from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.api import Api, Endpoints
from deepeval.dataset.utils import (
    convert_test_cases_to_goldens,
    convert_goldens_to_test_cases,
)
from deepeval.dataset.api import (
    APIDataset,
    CreateDatasetHttpResponse,
    Golden,
    DatasetHttpResponse,
)


@dataclass
class EvaluationDataset:
    test_cases: List[LLMTestCase]
    goldens: List[Golden]

    def __init__(self, test_cases: List[LLMTestCase] = []):
        self.test_cases = test_cases
        self.goldens = []

    def add_test_case(self, test_case: LLMTestCase):
        self.test_cases.append(test_case)

    def __iter__(self):
        return iter(self.test_cases)

    def evaluate(self, metrics: List[BaseMetric]):
        from deepeval import evaluate

        return evaluate(self.test_cases, metrics)

    def add_test_cases_from_csv_file(
        self,
        file_path: str,
        input_col_name: str,
        actual_output_col_name: str,
        expected_output_col_name: Optional[str] = None,
        context_col_name: Optional[str] = None,
        context_col_delimiter: str = ";",
    ):
        """
        Load test cases from a CSV file.

        This method reads a CSV file, extracting test case data based on specified column names. It creates LLMTestCase objects for each row in the CSV and adds them to the Dataset instance. The context data, if provided, is expected to be a delimited string in the CSV, which this method will parse into a list.

        Args:
            file_path (str): Path to the CSV file containing the test cases.
            input_col_name (str): The column name in the CSV corresponding to the input for the test case.
            actual_output_col_name (str): The column name in the CSV corresponding to the actual output for the test case.
            expected_output_col_name (str, optional): The column name in the CSV corresponding to the expected output for the test case. Defaults to None.
            context_col_name (str, optional): The column name in the CSV corresponding to the context for the test case. Defaults to None.
            context_delimiter (str, optional): The delimiter used to separate items in the context list within the CSV file. Defaults to ';'.

        Returns:
            None: The method adds test cases to the Dataset instance but does not return anything.

        Raises:
            FileNotFoundError: If the CSV file specified by `file_path` cannot be found.
            pd.errors.EmptyDataError: If the CSV file is empty.
            KeyError: If one or more specified columns are not found in the CSV file.

        Note:
            The CSV file is expected to contain columns as specified in the arguments. Each row in the file represents a single test case. The method assumes the file is properly formatted and the specified columns exist. For context data represented as lists in the CSV, ensure the correct delimiter is specified.
        """
        df = pd.read_csv(file_path)

        inputs = self._get_column_data(df, input_col_name)
        actual_outputs = self._get_column_data(df, actual_output_col_name)
        expected_outputs = self._get_column_data(
            df, expected_output_col_name, default=None
        )
        contexts = [
            context.split(context_col_delimiter) if context else []
            for context in self._get_column_data(
                df, context_col_name, default=""
            )
        ]

        for input, actual_output, expected_output, context in zip(
            inputs, actual_outputs, expected_outputs, contexts
        ):
            self.add_test_case(
                LLMTestCase(
                    input=input,
                    actual_output=actual_output,
                    expected_output=expected_output,
                    context=context,
                )
            )

    def _get_column_data(self, df: pd.DataFrame, col_name: str, default=None):
        return (
            df[col_name].values
            if col_name in df.columns
            else [default] * len(df)
        )

    def add_test_cases_from_json_file(
        self,
        file_path: str,
        input_key_name: str,
        actual_output_key_name: str,
        expected_output_key_name: Optional[str] = None,
        context_key_name: Optional[str] = None,
    ):
        """
        Load test cases from a JSON file.

        This method reads a JSON file containing a list of objects, each representing a test case. It extracts the necessary information based on specified key names and creates LLMTestCase objects to add to the Dataset instance.

        Args:
            file_path (str): Path to the JSON file containing the test cases.
            input_key_name (str): The key name in the JSON objects corresponding to the input for the test case.
            actual_output_key_name (str): The key name in the JSON objects corresponding to the actual output for the test case.
            expected_output_key_name (str, optional): The key name in the JSON objects corresponding to the expected output for the test case. Defaults to None.
            context_key_name (str, optional): The key name in the JSON objects corresponding to the context for the test case. Defaults to None.

        Returns:
            None: The method adds test cases to the Dataset instance but does not return anything.

        Raises:
            FileNotFoundError: If the JSON file specified by `file_path` cannot be found.
            ValueError: If the JSON file is not valid or if required keys (input and actual output) are missing in one or more JSON objects.

        Note:
            The JSON file should be structured as a list of objects, with each object containing the required keys. The method assumes the file format and keys are correctly defined and present.
        """
        try:
            with open(file_path, "r") as file:
                json_list = json.load(file)
        except FileNotFoundError:
            raise FileNotFoundError(f"The file {file_path} was not found.")
        except json.JSONDecodeError:
            raise ValueError(f"The file {file_path} is not a valid JSON file.")

        # Process each JSON object
        for json_obj in json_list:
            if (
                input_key_name not in json_obj
                or actual_output_key_name not in json_obj
            ):
                raise ValueError(
                    "Required fields are missing in one or more JSON objects"
                )

            input = json_obj[input_key_name]
            actual_output = json_obj[actual_output_key_name]
            expected_output = json_obj.get(expected_output_key_name)
            context = json_obj.get(context_key_name)

            self.add_test_case(
                LLMTestCase(
                    input=input,
                    actual_output=actual_output,
                    expected_output=expected_output,
                    context=context,
                )
            )

    def add_test_cases_from_hf_dataset(
        self,
        dataset_name: str,
        input_field_name: str,
        actual_output_field_name: str,
        expected_output_field_name: Optional[str] = None,
        context_field_name: Optional[str] = None,
        split: str = "train",
    ):
        """
        Load test cases from a Hugging Face dataset.

        This method loads a specified dataset and split from Hugging Face's datasets library, then iterates through each entry to create and add LLMTestCase objects to the Dataset instance based on specified field names.

        Args:
            dataset_name (str): The name of the Hugging Face dataset to load.
            split (str): The split of the dataset to load (e.g., 'train', 'test', 'validation'). Defaults to 'train'.
            input_field_name (str): The field name in the dataset corresponding to the input for the test case.
            actual_output_field_name (str): The field name in the dataset corresponding to the actual output for the test case.
            expected_output_field_name (str, optional): The field name in the dataset corresponding to the expected output for the test case. Defaults to None.
            context_field_name (str, optional): The field name in the dataset corresponding to the context for the test case. Defaults to None.

        Returns:
            None: The method adds test cases to the Dataset instance but does not return anything.

        Raises:
            ValueError: If the required fields (input and actual output) are not found in the dataset.
            FileNotFoundError: If the specified dataset is not available in Hugging Face's datasets library.
            datasets.DatasetNotFoundError: Specific Hugging Face error if the dataset or split is not found.
            json.JSONDecodeError: If there is an issue in reading or processing the dataset.

        Note:
            Ensure that the dataset structure aligns with the expected field names. The method assumes each dataset entry is a dictionary-like object.
        """

        try:
            from datasets import load_dataset
        except ImportError:
            raise ImportError(
                "The 'datasets' library is missing. Please install it using pip: pip install datasets"
            )
        hf_dataset = load_dataset(dataset_name, split=split)

        # Process each entry in the dataset
        for entry in hf_dataset:
            if (
                input_field_name not in entry
                or actual_output_field_name not in entry
            ):
                raise ValueError(
                    "Required fields are missing in one or more dataset entries"
                )

            input = entry[input_field_name]
            actual_output = entry[actual_output_field_name]
            expected_output = entry.get(expected_output_field_name)
            context = entry.get(context_field_name)

            self.add_test_case(
                LLMTestCase(
                    input=input,
                    actual_output=actual_output,
                    expected_output=expected_output,
                    context=context,
                )
            )

    def push(self, alias: str):
        if len(self.test_cases) == 0:
            raise ValueError(
                "Unable to push empty dataset to Confident AI, there must be at least one test case in dataset"
            )
        if os.path.exists(".deepeval"):
            goldens = convert_test_cases_to_goldens(self.test_cases)
            body = APIDataset(
                alias=alias, overwrite=False, goldens=goldens
            ).model_dump(by_alias=True, exclude_none=True)
            api = Api()
            result = api.post_request(
                endpoint=Endpoints.DATASET_ENDPOINT.value,
                body=body,
            )
            if result:
                response = CreateDatasetHttpResponse(
                    link=result["link"],
                )
                link = response.link
                console = Console()
                console.print(
                    "✅ Dataset successfully pushed to Confidnet AI! View at "
                    f"[link={link}]{link}[/link]"
                )
                webbrowser.open(link)
        else:
            raise Exception(
                "To push dataset to Confident AI, run `deepeval login`"
            )

    def pull(self, alias: str):
        if os.path.exists(".deepeval"):
            api = Api()
            result = api.get_request(
                endpoint=Endpoints.DATASET_ENDPOINT.value,
                params={"alias": alias},
            )
            response = DatasetHttpResponse(
                goldens=result["goldens"],
            )
            self.goldens.extend(response.goldens)

            # TODO: make this conversion at evaluation time instead
            self.test_cases.extend(convert_goldens_to_test_cases(self.goldens))
        else:
            raise Exception(
                "Run `deepeval login` to pull dataset from Confident AI"
            )



"""Tests for toxic
"""


import pytest
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.metrics import UnBiasedMetric
from deepeval import assert_test

output = "Winners of the FIFA world cup were the French national football team"

metric = UnBiasedMetric(evaluation_params=[LLMTestCaseParams.INPUT])


def test_non_bias_metric():
    metric = UnBiasedMetric(evaluation_params=[LLMTestCaseParams.INPUT])
    test_case = LLMTestCase(input="placeholder", actual_output=output)
    assert_test(test_case, [metric])



"""Utility function for getting API key.
"""
import os

from .constants import API_KEY_ENV
from .key_handler import KEY_FILE_HANDLER


def _get_api_key():
    """Get an API key here."""
    api_key = KEY_FILE_HANDLER.fetch_api_key()
    if api_key is None or api_key == "":
        api_key = os.getenv(API_KEY_ENV, "")
    return api_key



"""Tests for answer relevancy
"""
import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics import AnswerRelevancyMetric

from deepeval import assert_test, run_test

query = "What is Python?"
answer = "Python is a programming language?"


def test_query_answer_relevancy():
    scorer = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(input=query, actual_output=answer)
    assert_test(test_case, [scorer])


def test_compare_answer_relevancy_2():
    scorer = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(input=query, actual_output="Programming lang")
    run_test(test_case, metrics=[scorer])


def test_compare_answer_relevancy():
    metric = AnswerRelevancyMetric(minimum_score=0.5)
    query = "what is python"
    test_case_2 = LLMTestCase(
        input=query, actual_output="Python is a programming lang"
    )
    run_test(test_case_2, metrics=[metric])


def test_cross_encoder_answer_relevancy():
    metric = AnswerRelevancyMetric(
        minimum_score=0.5, model_type="cross_encoder"
    )
    test_case = LLMTestCase(input=query, actual_output=answer)
    assert_test(test_case, [metric])



from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCaseParams, LLMTestCase
from typing import List
from pydantic import BaseModel
from deepeval.api import Api


class JudgementalGPTResponse(BaseModel):
    score: float
    reason: str


class JudgementalGPTRequest(BaseModel):
    text: str
    criteria: str


class JudgementalGPT(BaseMetric):
    def __init__(
        self,
        name: str,
        criteria: str,
        evaluation_params: List[LLMTestCaseParams],
        minimum_score: float = 0.5,
    ):
        self.criteria = criteria
        self.name = name
        self.evaluation_params = evaluation_params
        self.minimum_score = minimum_score
        self.success = None
        self.reason = None

    @property
    def __name__(self):
        return self.name

    def measure(self, test_case: LLMTestCase):
        text = """"""
        for param in self.evaluation_params:
            value = getattr(test_case, param.value)
            text += f"{param.value}: {value} \n\n"

        judgemental_gpt_request_data = JudgementalGPTRequest(
            text=text, criteria=self.criteria
        )

        try:
            body = judgemental_gpt_request_data.model_dump(
                by_alias=True, exclude_none=True
            )
        except AttributeError:
            body = judgemental_gpt_request_data.dict(
                by_alias=True, exclude_none=True
            )
        api = Api()
        result = api.post_request(
            endpoint="/v1/judgemental-gpt",
            body=body,
        )
        response = JudgementalGPTResponse(
            score=result["score"],
            reason=result["reason"],
        )
        self.reason = response.reason
        self.score = response.score / 10
        self.success = self.score >= self.minimum_score

        return self.score

    def is_successful(self):
        return self.success



import os

import pytest
from deepeval.dataset import EvaluationDataset
from deepeval.metrics import HallucinationMetric
from deepeval import assert_test
from deepeval.test_case import LLMTestCase

dataset = EvaluationDataset()


def test_create_dataset():
    module_b_dir = os.path.dirname(os.path.realpath(__file__))

    file_path = os.path.join(module_b_dir, "data", "dataset.csv")

    dataset.add_test_cases_from_csv_file(
        file_path,
        input_col_name="query",
        actual_output_col_name="actual_output",
        expected_output_col_name="expected_output",
        context_col_name="context",
    )
    assert len(dataset.test_cases) == 5, "Test Cases not loaded from CSV"
    file_path = os.path.join(module_b_dir, "data", "dataset.json")
    dataset.add_test_cases_from_json_file(
        file_path,
        input_key_name="query",
        actual_output_key_name="actual_output",
        expected_output_key_name="expected_output",
        context_key_name="context",
    )
    assert len(dataset.test_cases) == 10, "Test Cases not loaded from JSON"

    # dataset.push("alias")


# dataset.pull("alias")
# @pytest.mark.parametrize(
#     "test_case",
#     dataset,
# )
# def test_customer_chatbot(test_case: LLMTestCase):
#     hallucination_metric = HallucinationMetric(minimum_score=0.3)
#     assert_test(test_case, [hallucination_metric])



from setuptools import find_packages, setup

from deepeval._version import __version__

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()


setup(
    name="deepeval",
    version=__version__,
    url="https://github.com/confident-ai.com/deepeval",
    author="Confident AI",
    author_email="jeffreyip@confident-ai.com",
    description="The open-source evaluation framework for LLMs.",
    packages=find_packages(),
    # TODO - make pandas an 'extra' requirement in the future
    install_requires=[
        "requests",
        "tqdm",
        "transformers",
        "pytest",
        "tabulate",
        "sentence-transformers",  # for similarity
        "pytest",
        "typer==0.9.0",
        "rich",
        "protobuf>=4.21.6",
        "pandas",
        "pydantic",  # loosen pydantic requirements as we support multiple
        "sentry-sdk",
        "pytest-xdist",
        "portalocker",
        "langchain",
        "rouge_score==0.1.2",
        "nltk==3.8.1",
        "ragas",
        "detoxify",
    ],
    extras_require={
        "bias": [
            "tensorflow",  # for bias
            "Dbias",  # for bias
        ],
        "dev": ["black"],
    },
    entry_points={
        "console_scripts": [
            "deepeval = deepeval.cli.main:app",
        ],
        "pytest11": [
            "plugins = deepeval.plugins.plugin",
        ],
    },
)



from enum import Enum
import os
import time
from typing import Any
import tqdm
import re
import string
import numpy as np
from dataclasses import asdict, is_dataclass


def dataclass_to_dict(instance: Any) -> Any:
    if is_dataclass(instance):
        return {k: dataclass_to_dict(v) for k, v in asdict(instance).items()}
    elif isinstance(instance, Enum):
        return instance.value
    elif isinstance(instance, list):
        return [dataclass_to_dict(item) for item in instance]
    elif isinstance(instance, tuple):
        return tuple(dataclass_to_dict(item) for item in instance)
    elif isinstance(instance, dict):
        return {k: dataclass_to_dict(v) for k, v in instance.items()}
    else:
        return instance


def delete_file_if_exists(file_path):
    try:
        if os.path.exists(file_path):
            os.remove(file_path)
    except Exception as e:
        print(f"An error occurred: {e}")


def softmax(x):
    e_x = np.exp(x - np.max(x, axis=1, keepdims=True))
    return e_x / e_x.sum(axis=1, keepdims=True)


def cosine_similarity(vector_a, vector_b):
    dot_product = np.dot(vector_a, vector_b)
    norm_a = np.linalg.norm(vector_a)
    norm_b = np.linalg.norm(vector_b)
    similarity = dot_product / (norm_a * norm_b)
    return similarity


def chunk_text(text, chunk_size=20):
    words = text.split()
    chunks = [
        " ".join(words[i : i + chunk_size])
        for i in range(0, len(words), chunk_size)
    ]
    return chunks


def normalize_text(text: str) -> str:
    """Lower text and remove punctuation, articles and extra whitespace.
    Copied from the [QuAC](http://quac.ai/) evaluation script found at
    https://s3.amazonaws.com/my89public/quac/scorer.py"""

    def remove_articles(text: str) -> str:
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text: str) -> str:
        return " ".join(text.split())

    def remove_punc(text: str) -> str:
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text: str) -> str:
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(text))))


###############################################
# Source: https://github.com/tingofurro/summac
###############################################

# GPU-related business


def get_freer_gpu():
    os.system("nvidia-smi -q -d Memory |grep -A4 GPU|grep Free >tmp_smi")
    memory_available = [
        int(x.split()[2]) + 5 * i
        for i, x in enumerate(open("tmp_smi", "r").readlines())
    ]
    os.remove("tmp_smi")
    return np.argmax(memory_available)


def any_gpu_with_space(gb_needed):
    os.system("nvidia-smi -q -d Memory |grep -A4 GPU|grep Free >tmp_smi")
    memory_available = [
        float(x.split()[2]) / 1024.0
        for i, x in enumerate(open("tmp_smi", "r").readlines())
    ]
    os.remove("tmp_smi")
    return any([mem >= gb_needed for mem in memory_available])


def wait_free_gpu(gb_needed):
    while not any_gpu_with_space(gb_needed):
        time.sleep(30)


def select_freer_gpu():
    freer_gpu = str(get_freer_gpu())
    print("Will use GPU: %s" % (freer_gpu))
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    os.environ["CUDA_VISIBLE_DEVICES"] = "" + freer_gpu
    return freer_gpu


def batcher(iterator, batch_size=4, progress=False):
    if progress:
        iterator = tqdm.tqdm(iterator)

    batch = []
    for elem in iterator:
        batch.append(elem)
        if len(batch) == batch_size:
            final_batch = batch
            batch = []
            yield final_batch
    if len(batch) > 0:  # Leftovers
        yield batch
